from ..solve import solve
from ..liealgebra import seLieAlgebra
from ..liegroup import SELieGroup
import numpy as np
import unittest


def spinning_top(
    t, y, principal_moments=np.array([2, 2, 1]), m=1, g=1, chi=np.array([0, 0, 1])
):
    """A formulation of the problem exploiting the Lie-group structure"""
    mu, beta = np.split(y, 2)
    mu_dot = -mu / principal_moments
    beta_dot = -m * g * chi
    return np.hstack((mu_dot, beta_dot))


se3 = seLieAlgebra(SELieGroup())


class TestHeavyTop(unittest.TestCase):
    def test_solve(self):
        y0 = np.array([np.sin(1.1), 0, np.cos(1.1), 1, 0.2, 3])
        t_start = 0
        t_end = 5
        step_length = 0.01
        manifold = "heavytop"
        method = "RKMK4"
        solution = solve(
            spinning_top, y0, t_start, t_end, step_length, manifold, method
        )  # noqa: E501
        expected_norm = np.linalg.norm(y0[3:])
        for i in range(len(solution.T)):
            self.assertAlmostEqual(np.linalg.norm(solution[3:, i]), expected_norm)

    def test_dexpinv_zero(self):
        u = np.zeros(6)
        v = np.random.random(6)
        np.testing.assert_array_equal(se3.dexpinv(u, v), v)
        np.testing.assert_array_equal(se3.dexpinv(v, u), u)

    def test_dexpinv_nonzero(self):
        u = np.array([0.89120736, 0.0, 0.45359612, 1.0, 0.2, 3.0])
        v = np.array(
            [0.98688917, 0.62018318, 0.66178247, 0.67912063, 0.43851834, 0.8804615]
        )
        expected = np.array(
            [
                1.133009950431,
                0.638687002977,
                0.374690254641,
                1.618548639942,
                -0.730601007173,
                0.603188282585,
            ]
        )
        actual = se3.dexpinv(u, v)
        np.testing.assert_array_almost_equal(actual, expected)


if __name__ == "__main__":
    unittest.main()



import numpy as np


class LieGroup:
    def __init__(self):
        pass

    def action(self, g, u):
        raise NotImplementedError


class SOLieGroup(LieGroup):
    def action(self, g, u):
        return g @ u


class SELieGroup(LieGroup):
    def action(self, g, u):
        """The left (coadjoint) action of the
        group SE(3) on the dual of its lie algebra se(3)*

        Parameters
        ----------
        g : Two-tuple
            Element of the Lie group. First element is a 3x3 matrix,
            second element a 3-vector.
        u : array of length 6
            Element of the dual of se(3).

        Returns
        -------
        [type]
            [description]
        """
        G, g = g
        u, v = np.split(u, 2)
        z2 = G @ v
        z1 = G @ u + np.cross(g, z2)
        return np.hstack((z1, z2))



import numpy as np
from scipy.linalg import expm


class LieAlgebra:
    def __init__(self, LieGroup) -> None:
        self.action = LieGroup.action

    def exp(self, y):
        return expm(y)

    def dexpinv(self, u, v, order: int):
        ans = v
        if order >= 2:
            c = self.commutator(u, v)
            ans -= (1 / 2) * c
        if order >= 4:
            c = self.commutator(u, c)
            ans += (1 / 12) * c
        if order >= 6:
            raise NotImplementedError("Not yet implemented for order >= 6")
        return ans

    def commutator(self, a, b):
        if a.ndim == 2 and b.ndim == 2 and a.shape == b.shape:
            # a and b are matrices of similar dimensions
            return a @ b - b @ a
        else:
            raise NotImplementedError


class soLieAlgebra(LieAlgebra):
    def exp(self, y):
        if y.size == 3 and y.ndim == 1:
            # Use the rodrigues formula

            # Check if y is the zero vector
            if np.array_equal(y, np.zeros(3)):
                return np.eye(3)

            alpha = np.linalg.norm(y)
            Y = self.matrix(y)
            return (
                np.eye(3)
                + (np.sin(alpha) / alpha) * Y
                + ((1 - np.cos(alpha)) / alpha ** 2) * Y @ Y
            )
        # Otherwise, use the standard expm formula
        # We are here assuming that y is a matrix
        return super().exp(y)

    def dexpinv(self, u, v, _):
        if u.size == 3 and u.ndim == 1:
            # Use Rodrigues formula

            # Convert v from the matrix representation to the basis representation
            v = np.array([v[2, 1], v[0, 2], v[1, 0]])

            # Check for the zero vector
            if np.array_equal(u, np.zeros(3)):
                return v

            cot = lambda x: 1 / np.tan(x)  # noqa: E731

            alpha = np.linalg.norm(u)
            U = self.matrix(u)

            return self.action(
                np.eye(3)
                - 0.5 * U
                - (2 - alpha * cot(0.5 * alpha)) / (2 * alpha ** 2) * U @ U,
                v,
            )
        else:
            return super().dexpinv(u, v)

    def matrix(self, y):
        if y.size != 3:
            raise NotImplementedError("Not yet implemented for n != 3")
        u, v, w = y
        return np.array([[0, -w, v], [w, 0, -u], [-v, u, 0]])


class seLieAlgebra(LieAlgebra):
    def _hat(self, y):
        u, v, w = y
        return np.array([[0, -w, v], [w, 0, -u], [-v, u, 0]])

    def _inv_hat(self, y):
        return np.array([y[2, 1], y[0, 2], y[1, 0]])

    def exp(self, y):
        u, v = np.split(y, 2)
        alpha = np.linalg.norm(u)
        u_hat = self._hat(u)
        if np.isclose(alpha, 0):
            u_exp = np.eye(3)
            v_exp = v
        else:
            u_exp = (
                np.eye(3)
                + (np.sin(alpha) / alpha) * u_hat
                + ((1 - np.cos(alpha)) / alpha ** 2) * u_hat @ u_hat
            )
            v_exp = (
                np.eye(3)
                + (1 - np.cos(alpha)) / (alpha ** 2) * u_hat
                + (alpha - np.sin(alpha)) / (alpha ** 3) * u_hat @ u_hat
            ) @ v
        return (u_exp, v_exp)

    def _cot(self, x):
        return 1 / np.tan(x)

    def _csc(self, x):
        return 1 / np.sin(x)

    def _dexpinv_helper_1(self, z):
        return (1 - 0.5 * z * self._cot(0.5 * z)) / (z ** 2)

    def _dexpinv_helper_2(self, z, rho):
        return (
            0.25
            * rho
            * ((z * self._csc(0.5 * z)) ** 2 + 2 * z * self._cot(0.5 * z) - 8)
            / (z ** 4)
        )

    def dexpinv(self, u, v, _=None):
        """Returns the result of dexp^(-1)_(u) (v).
        Both u and v are elements of se(3), represented as vectors of length 6"""
        A, a = np.split(u, 2)
        B, b = np.split(v, 2)
        alpha = np.linalg.norm(A)
        rho = np.inner(A, a)
        if np.isclose(alpha, 0):
            return v
        c1 = (
            B
            - 0.5 * np.cross(A, B)
            + self._dexpinv_helper_1(alpha) * np.cross(A, np.cross(A, B))
        )
        c2 = (
            b
            - 0.5 * (np.cross(a, B) + np.cross(A, b))
            + self._dexpinv_helper_2(alpha, rho) * np.cross(A, np.cross(A, B))
            + self._dexpinv_helper_1(alpha)
            * (
                np.cross(a, np.cross(A, B))
                + np.cross(A, np.cross(a, B))
                + np.cross(A, np.cross(A, b))
            )
        )
        return np.hstack((c1, c2))



import numpy as np
from ..liealgebra import seLieAlgebra
from ..liegroup import SELieGroup
from .hmanifold import HomogenousManifold


class HeavyTop(HomogenousManifold):
    def __init__(self, y):
        if not isinstance(y, np.ndarray):
            try:
                y = np.array(y)
            except Exception:
                raise TypeError("y must be array_like")
        self.n = y.size
        self.y = y
        self.lie_group = SELieGroup()
        self.lie_algebra = seLieAlgebra(self.lie_group)
        super().__init__()
